import jax
import navix as nx


def test_tasks_composition():
    reward_fn = nx.tasks.compose(
        nx.tasks.navigation,
        nx.tasks.action_cost,
        nx.tasks.time_cost,
        nx.tasks.wall_hit_cost,
    )

    env = nx.environments.Room(height=3, width=3, max_steps=8, reward_fn=reward_fn)
    key = jax.random.PRNGKey(0)

    def _test():
        timestep = env.reset(key)
        for _ in range(10):
            timestep = env.step(timestep, jax.random.randint(key, (), 0, 7))
        return timestep

    print(jax.jit(_test)())


if __name__ == "__main__":
    test_tasks_composition()



from __future__ import annotations
import dataclasses
from typing import Dict, Tuple, Type, TypeVar

import jax
from jax import Array
import jax.numpy as jnp
from flax import struct
from jax.random import KeyArray
from jax_enums import Enumerable

from .components import (
    Positionable,
    Directional,
    HasTag,
    Stochastic,
    Openable,
    Pickable,
    Holder,
    HasSprite,
)
from .graphics import RenderingCache, SPRITES_REGISTRY
from .config import config

T = TypeVar("T", bound="Entity")


class Entities(struct.PyTreeNode):
    WALL: str = struct.field(pytree_node=False, default="wall")
    FLOOR: str = struct.field(pytree_node=False, default="floor")
    PLAYER: str = struct.field(pytree_node=False, default="player")
    GOAL: str = struct.field(pytree_node=False, default="goal")
    KEY: str = struct.field(pytree_node=False, default="key")
    DOOR: str = struct.field(pytree_node=False, default="door")


class Entity(Positionable, HasTag, HasSprite):
    """Entities are components that can be placed in the environment"""

    def __post_init__(self) -> None:
        if not config.ARRAY_CHECKS_ENABLED:
            return
        # Check that all fields have the same batch size
        fields = self.__dataclass_fields__
        batch_size = self.shape[0:]
        for path, leaf in jax.tree_util.tree_leaves_with_path(self):
            name = path[0].name
            default_ndim = len(fields[name].metadata["shape"])
            prefix = int(default_ndim != leaf.ndim)
            leaf_batch_size = leaf.shape[:prefix]
            assert (
                leaf_batch_size == batch_size
            ), f"Expected {name} to have batch size {batch_size}, got {leaf_batch_size} instead"

    def check_ndim(self, batched: bool = False) -> None:
        if not config.ARRAY_CHECKS_ENABLED:
            return
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            default_ndim = len(field.metadata["shape"])
            assert (
                value.ndim == default_ndim + batched
            ), f"Expected {field.name} to have ndim {default_ndim - batched}, got {value.ndim} instead"

    def __getitem__(self: T, idx) -> T:
        return jax.tree_util.tree_map(lambda x: x[idx], self)

    @property
    def shape(self) -> Tuple[int, ...]:
        """The batch shape of the entity"""
        return self.position.shape[: self.position.ndim - 1]

    @property
    def walkable(self) -> Array:
        raise NotImplementedError()

    @property
    def transparent(self) -> Array:
        raise NotImplementedError()


class Wall(Entity):
    """Walls are entities that cannot be walked through"""

    @classmethod
    def create(
        cls,
        position: Array,
    ) -> Wall:
        return cls(position=position)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.WALL]
        return jnp.broadcast_to(sprite[None], (*self.shape, *sprite.shape))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(0), self.shape)


class Player(Entity, Directional, Holder):
    """Players are entities that can act around the environment"""

    @classmethod
    def create(
        cls,
        position: Array,
        direction: Array,
        pocket: Array,
    ) -> Player:
        return cls(position=position, direction=direction, pocket=pocket)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.PLAYER][self.direction]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # broadcast to batch_size
        return jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(2), self.shape)


class Goal(Entity, Stochastic):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        probability: Array,
    ) -> Goal:
        return cls(position=position, probability=probability)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.GOAL]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(3), self.shape)


class Key(Entity, Pickable):
    """Pickable items are world objects that can be picked up by the player.
    Examples of pickable items are keys, coins, etc."""

    @classmethod
    def create(
        cls,
        position: Array,
        id: Array,
    ) -> Key:
        return cls(position=position, id=id)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.KEY]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(4), self.shape)


class Door(Entity, Directional, Openable):
    """Consumable items are world objects that can be consumed by the player.
    Consuming an item requires a tool (e.g. a key to open a door).
    A tool is an id (int) of another item, specified in the `requires` field (-1 if no tool is required).
    After an item is consumed, it is both removed from the `state.entities` collection, and replaced in the grid
    by the item specified in the `replacement` field (0 = floor by default).
    Examples of consumables are doors (to open) food (to eat) and water (to drink), etc.
    """

    @classmethod
    def create(
        cls,
        position: Array,
        direction: Array,
        requires: Array,
        open: Array,
    ) -> Door:
        return cls(position=position, direction=direction, requires=requires, open=open)

    @property
    def walkable(self) -> Array:
        return self.open

    @property
    def transparent(self) -> Array:
        return self.open

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.DOOR][
            self.direction, jnp.asarray(self.open, dtype=jnp.int32)
        ]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(5), self.shape)


class State(struct.PyTreeNode):
    """The Markovian state of the environment"""

    key: KeyArray
    """The random number generator state"""
    grid: Array
    """The base map of the environment that remains constant throughout the training"""
    cache: RenderingCache
    """The rendering cache to speed up rendering"""
    entities: Dict[str, Entity] = struct.field(default_factory=dict)
    """The entities in the environment, indexed via entity type string representation.
    Batched over the number of entities for each type"""

    def get_entity(self, entity_enum: str) -> Entity:
        return self.entities[entity_enum]

    def set_entity(self, entity_enum: str, entity: Entity) -> State:
        self.entities[entity_enum] = entity
        return self

    def get_walls(self) -> Wall:
        return self.entities.get(Entities.WALL, Wall())  # type: ignore

    def set_walls(self, walls: Wall) -> State:
        self.entities[Entities.WALL] = walls
        return self

    def get_player(self, idx: int = 0) -> Player:
        return self.entities[Entities.PLAYER][idx]  # type: ignore

    def set_player(self, player: Player, idx: int = 0) -> State:
        # TODO(epignatelli): this is a hack and won't work in multi-agent settings
        self.entities[Entities.PLAYER] = player[None]
        return self

    def get_goals(self) -> Goal:
        return self.entities[Entities.GOAL]  # type: ignore

    def set_goals(self, goals: Goal) -> State:
        self.entities[Entities.GOAL] = goals
        return self

    def get_keys(self) -> Key:
        return self.entities[Entities.KEY]  # type: ignore

    def set_keys(self, keys: Key) -> State:
        self.entities[Entities.KEY] = keys
        return self

    def get_doors(self) -> Door:
        return self.entities[Entities.DOOR]  # type: ignore

    def set_doors(self, doors: Door) -> State:
        self.entities[Entities.DOOR] = doors
        return self

    def get_positions(self) -> Array:
        return jnp.concatenate([self.entities[k].position for k in self.entities])

    def get_tags(self) -> Array:
        return jnp.concatenate([self.entities[k].tag for k in self.entities])

    def get_sprites(self) -> Array:
        return jnp.concatenate([self.entities[k].sprite for k in self.entities])

    def get_transparency(self) -> Array:
        return jnp.concatenate([self.entities[k].transparent for k in self.entities])



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from jax import Array
import jax.numpy as jnp
from .entities import State
from .grid import positions_equal


def check_truncation(terminated: Array, truncated: Array) -> Array:
    result = jnp.asarray(truncated + 2 * terminated, dtype=jnp.int32)
    return jnp.clip(result, 0, 2)


def on_navigation_completion(prev_state: State, action: Array, state: State) -> Array:
    player = state.get_player()
    goals = state.get_goals()

    reached = positions_equal(player.position, goals.position)
    return jnp.any(reached)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Union

import jax
import jax.numpy as jnp
from jax.random import KeyArray

from ..components import EMPTY_POCKET_ID
from ..entities import Entities, Goal, Player, State
from ..grid import random_positions, random_directions, room
from ..graphics import RenderingCache
from .environment import Environment, Timestep


class Room(Environment):
    def reset(
        self, key: KeyArray, cache: Union[RenderingCache, None] = None
    ) -> Timestep:
        key, k1, k2 = jax.random.split(key, 3)

        # map
        grid = room(height=self.height, width=self.width)

        # player
        player_pos, goal_pos = random_positions(k1, grid, n=2)
        direction = random_directions(k2, n=1)
        player = Player(
            position=player_pos,
            direction=direction,
            pocket=EMPTY_POCKET_ID,
        )
        # goal
        goal = Goal(position=goal_pos, probability=jnp.asarray(1.0))

        entities = {
            Entities.PLAYER: player[None],
            Entities.GOAL: goal[None],
        }

        # systems
        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )

        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation(state),
            action=jnp.asarray(0, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )



from __future__ import annotations

from typing import Any, Dict, Tuple

import jax
import jax.numpy as jnp
from flax import struct
from jax import Array


TILE_SIZE = 32
RED = jnp.asarray([255, 0, 0], dtype=jnp.uint8)
GREEN = jnp.asarray([0, 255, 0], dtype=jnp.uint8)
BLUE = jnp.asarray([0, 0, 255], dtype=jnp.uint8)
BLACK = jnp.asarray([0, 0, 0], dtype=jnp.uint8)
WHITE = jnp.asarray([255, 255, 255], dtype=jnp.uint8)
YELLOW = jnp.asarray([255, 255, 0], dtype=jnp.uint8)
PURPLE = jnp.asarray([255, 0, 255], dtype=jnp.uint8)
CYAN = jnp.asarray([0, 255, 255], dtype=jnp.uint8)
ORANGE = jnp.asarray([255, 128, 0], dtype=jnp.uint8)
PINK = jnp.asarray([255, 0, 128], dtype=jnp.uint8)
BROWN = jnp.asarray([128, 64, 0], dtype=jnp.uint8)
GRAY_10 = jnp.asarray([230, 230, 230], dtype=jnp.uint8)
GRAY_20 = jnp.asarray([205, 205, 205], dtype=jnp.uint8)
GRAY_40 = jnp.asarray([153, 153, 153], dtype=jnp.uint8)
GRAY_50 = jnp.asarray([128, 128, 128], dtype=jnp.uint8)
GRAY_70 = jnp.asarray([77, 77, 77], dtype=jnp.uint8)
GRAY_80 = jnp.asarray([51, 51, 51], dtype=jnp.uint8)
GRAY_90 = jnp.asarray([25, 25, 25], dtype=jnp.uint8)
GOLD = jnp.asarray([255, 215, 0], dtype=jnp.uint8)
SILVER = jnp.asarray([192, 192, 192], dtype=jnp.uint8)
BRONZE = jnp.asarray([205, 127, 50], dtype=jnp.uint8)
MAROON = jnp.asarray([128, 0, 0], dtype=jnp.uint8)
NAVY = jnp.asarray([0, 0, 128], dtype=jnp.uint8)
TEAL = jnp.asarray([0, 128, 128], dtype=jnp.uint8)
OLIVE = jnp.asarray([128, 128, 0], dtype=jnp.uint8)
LIME = jnp.asarray([0, 255, 0], dtype=jnp.uint8)
AQUA = jnp.asarray([0, 255, 255], dtype=jnp.uint8)
FUCHSIA = jnp.asarray([255, 0, 255], dtype=jnp.uint8)
SALMON = jnp.asarray([250, 128, 114], dtype=jnp.uint8)
TURQUOISE = jnp.asarray([64, 224, 208], dtype=jnp.uint8)
VIOLET = jnp.asarray([238, 130, 238], dtype=jnp.uint8)
INDIGO = jnp.asarray([75, 0, 130], dtype=jnp.uint8)
BEIGE = jnp.asarray([245, 245, 220], dtype=jnp.uint8)
MINT = jnp.asarray([189, 252, 201], dtype=jnp.uint8)
LAVENDER = jnp.asarray([230, 230, 250], dtype=jnp.uint8)
APRICOT = jnp.asarray([251, 206, 177], dtype=jnp.uint8)
MAUVE = jnp.asarray([224, 176, 255], dtype=jnp.uint8)
LILAC = jnp.asarray([200, 162, 200], dtype=jnp.uint8)
TAN = jnp.asarray([210, 180, 140], dtype=jnp.uint8)


def colour_chart(size: int = TILE_SIZE) -> Array:
    colours = [
        RED,
        GREEN,
        BLUE,
        BLACK,
        WHITE,
        YELLOW,
        PURPLE,
        CYAN,
        ORANGE,
        PINK,
        BROWN,
        GRAY_20,
        GRAY_50,
        GRAY_70,
        GRAY_90,
        GOLD,
        SILVER,
        BRONZE,
        MAROON,
        NAVY,
        TEAL,
        OLIVE,
        LIME,
        AQUA,
        FUCHSIA,
        SALMON,
        TURQUOISE,
        VIOLET,
        INDIGO,
        BEIGE,
        MINT,
        LAVENDER,
        APRICOT,
        MAUVE,
        LILAC,
        TAN,
    ]
    grid = jnp.zeros((size * len(colours), size * len(colours), 3), dtype=jnp.uint8)
    for i, colour in enumerate(colours):
        for j, colour in enumerate(colours):
            grid = grid.at[i * size : (i + 1) * size, j * size : (j + 1) * size].set(
                colour
            )
    return grid


class RenderingCache(struct.PyTreeNode):
    patches: Array
    """A flat set of patches representing the RGB values of each tile in the base map"""

    @classmethod
    def init(cls, grid: Array) -> RenderingCache:
        background = render_background(grid)
        patches = flatten_patches(background)

        # add discard pile
        patches = jnp.concatenate(
            [
                patches,
                jnp.zeros((1, TILE_SIZE, TILE_SIZE, 3), dtype=jnp.uint8),
            ],
            axis=0,
        )
        return cls(patches=patches)


def colorise_tile(tile: Array, colour: Array, background: Array = WHITE) -> Array:
    assert tile.shape == (
        TILE_SIZE,
        TILE_SIZE,
    ), "Tile must be of size TILE_SIZE, TILE_SIZE, 3, got {}".format(tile.shape)
    tile = jnp.stack([tile] * colour.shape[0], axis=-1)
    tile = jnp.where(tile, colour, background)
    return tile


def render_rectangle(size: int = TILE_SIZE, colour: Array = BLACK) -> Array:
    rectangle = jnp.ones((size - 2, size - 2), dtype=jnp.int32)
    rectangle = jnp.pad(rectangle, 1, "constant", constant_values=0)
    return colorise_tile(rectangle, colour)


def render_triangle_east(size: int = TILE_SIZE, colour: Array = RED) -> Array:
    triangle = jnp.ones((size, size), dtype=jnp.int32)
    triangle = jnp.tril(triangle, k=0)
    triangle = jnp.flip(triangle, axis=0)
    triangle = jnp.tril(triangle, k=0)
    triangle = jnp.roll(triangle, (0, size // 3))
    return colorise_tile(triangle, colour)


def render_triangle_south(size: int = TILE_SIZE, colour: Array = RED) -> Array:
    triangle = render_triangle_east(size)
    triangle = jnp.rot90(triangle, k=3)
    return colorise_tile(triangle, colour)


def render_triangle_west(size: int = TILE_SIZE, colour: Array = RED) -> Array:
    triangle = render_triangle_east(size)
    triangle = jnp.rot90(triangle, k=2)
    return colorise_tile(triangle, colour)


def render_triangle_north(size: int = TILE_SIZE, colour: Array = RED) -> Array:
    triangle = render_triangle_east(size)
    triangle = jnp.rot90(triangle, k=1)
    return colorise_tile(triangle, colour)


def render_diamond(size: int = TILE_SIZE, colour: Array = GOLD) -> Array:
    diamond = jnp.ones((size, size), dtype=jnp.int32)
    diamond = jnp.tril(diamond, k=size // 2.5)
    diamond = jnp.flip(diamond, axis=0)
    diamond = jnp.tril(diamond, k=size // 2.5)
    diamond = jnp.flip(diamond, axis=1)
    diamond = jnp.tril(diamond, k=size // 2.5)
    diamond = jnp.flip(diamond, axis=0)
    diamond = jnp.tril(diamond, k=size // 2.5)
    return colorise_tile(diamond, colour)


def render_door_closed(size: int = TILE_SIZE, colour: Array = BROWN) -> Array:
    frame_size = size - 6
    door = jnp.zeros((frame_size, frame_size), dtype=jnp.int32)
    door = jnp.pad(door, 1, "constant", constant_values=1)
    door = jnp.pad(door, 1, "constant", constant_values=0)
    door = jnp.pad(door, 1, "constant", constant_values=1)

    x_0 = size - size // 4
    y_centre = size // 2
    y_size = size // 5
    door = door.at[y_centre - y_size // 2 : y_centre + y_size // 2, x_0 : x_0 + 1].set(
        1
    )
    return colorise_tile(door, colour)


def render_door_locked(size: int = TILE_SIZE, colour: Array = BROWN) -> Array:
    frame_size = size - 4
    door = jnp.zeros((frame_size, frame_size), dtype=jnp.int32)
    door = jnp.pad(door, 2, "constant", constant_values=1)

    x_0 = size - size // 4
    y_centre = size // 2
    y_size = size // 5
    door = door.at[y_centre - y_size // 2 : y_centre + y_size // 2, x_0 : x_0 + 1].set(
        1
    )
    return colorise_tile(door, colour, background=colour / 2)


def render_door_open(size: int = TILE_SIZE, colour: Array = BROWN) -> Array:
    door = jnp.zeros((size, size), dtype=jnp.int32)
    door = door.at[0].set(1)
    door = door.at[3].set(1)
    door = door.at[:3, 0].set(1)
    door = door.at[:3, -1].set(1)
    return colorise_tile(door, colour)


def render_key(size: int = TILE_SIZE, colour: Array = BRONZE) -> Array:
    key = jnp.zeros((size, size), dtype=jnp.int32)

    # Handle (Round Part)
    handle_radius = size // 4
    handle_center = (size // 2, size // 4)
    y, x = jnp.ogrid[:size, :size]
    mask = (x - handle_center[0]) ** 2 + (
        y - handle_center[1]
    ) ** 2 <= handle_radius**2
    key = jnp.where(mask, 1, key)

    # Shaft (Straight Part)
    shaft_width = size // 8
    shaft_height = size // 2
    shaft_start = (size // 2 - shaft_width // 2, size // 2 - shaft_height // 2)
    shaft_end = (size // 2 + shaft_width // 2, size // 2 + shaft_height // 2)
    shaft_mask = jnp.logical_and(
        jnp.logical_and(x >= shaft_start[0], x <= shaft_end[0]),
        jnp.logical_and(y >= shaft_start[1], y <= shaft_end[1]),
    )
    key = jnp.where(shaft_mask, 1, key)

    # Tooth (Pointy End)
    tooth_width = size // 15
    tooth_height = size // 2
    tooth_position = (size // 2 - tooth_width // 2, size - tooth_height)
    tooth_mask = jnp.logical_and(
        jnp.logical_and(x >= tooth_position[0], x <= tooth_position[0] + tooth_width),
        jnp.logical_and(y >= tooth_position[1], y <= tooth_position[1] + tooth_height),
    )
    key = jnp.where(tooth_mask, 1, key)

    return colorise_tile(key, colour)


def render_floor(size: int = TILE_SIZE, colour: Array = WHITE) -> Array:
    floor = jnp.ones((size - 2, size - 2), dtype=jnp.int32)
    floor = jnp.pad(floor, 1, "constant", constant_values=0)
    return colorise_tile(floor, colour, background=GRAY_10)


def render_wall(size: int = TILE_SIZE, colour: Array = GRAY_80) -> Array:
    wall = jnp.ones((size, size), dtype=jnp.int32)
    return colorise_tile(wall, colour)


def tile_grid(grid: Array, tile: Array) -> Array:
    tiled = jnp.tile(tile, (*grid.shape, 1))
    return jnp.asarray(tiled, dtype=jnp.uint8)


@jax.jit
def build_sprites_registry() -> Dict[str, Any]:
    registry = {}

    wall = render_wall()
    floor = render_floor()
    player = render_triangle_east()
    goal = render_diamond()
    key = render_key()
    door_closed = render_door_closed()
    door_open = render_door_open()

    # 0: set wall sprites
    registry["wall"] = wall

    # 1: set floor sprites
    registry["floor"] = floor

    # 2: set player sprites
    registry["player"] = jnp.stack(
        [
            player,
            jnp.rot90(player, k=3),
            jnp.rot90(player, k=2),
            jnp.rot90(player, k=1),
        ]
    )

    # 3: set goal sprites
    registry["goal"] = goal

    # 4: set key sprites
    registry["key"] = key

    # 5: set door sprites
    door = jnp.zeros((4, 2, TILE_SIZE, TILE_SIZE, 3), dtype=jnp.uint8)

    door_closed_by_direction = jnp.stack(
        [
            jnp.rot90(door_closed, k=1),
            door_closed,
            jnp.rot90(door_closed, k=3),
            jnp.rot90(door_closed, k=2),
        ]
    )
    door = door.at[:, 0].set(door_closed_by_direction)

    door_open_by_direction = jnp.stack(
        [
            door_open,
            jnp.rot90(door_open, k=1),
            jnp.rot90(door_open, k=2),
            jnp.rot90(door_open, k=3),
        ]
    )
    door = door.at[:, 1].set(door_open_by_direction)

    registry["door"] = door

    return registry


SPRITES_REGISTRY: Dict[str, Any] = build_sprites_registry()


def render_background(
    grid: Array, sprites_registry: Dict[str, Any] = SPRITES_REGISTRY
) -> Array:
    image_width = grid.shape[0] * TILE_SIZE
    image_height = grid.shape[1] * TILE_SIZE
    n_channels = 3

    background = jnp.zeros((image_height, image_width, n_channels), dtype=jnp.uint8)
    grid_resized = jax.image.resize(
        grid, (grid.shape[0] * TILE_SIZE, grid.shape[1] * TILE_SIZE), method="nearest"
    )

    mask = jnp.asarray(grid_resized, dtype=bool)  # 0 = floor, 1 = wall
    # index by [entity_type, direction, open/closed, y, x, channel]
    wall_tile = tile_grid(grid, sprites_registry["wall"])
    floor_tile = tile_grid(grid, sprites_registry["floor"])
    background = jnp.where(mask[..., None], wall_tile, floor_tile)
    return background


def flatten_patches(
    image: Array, patch_size: Tuple[int, int] = (TILE_SIZE, TILE_SIZE)
) -> Array:
    height = image.shape[0] // patch_size[0]
    width = image.shape[1] // patch_size[1]
    n_channels = image.shape[2]

    grid = image.reshape(height, patch_size[0], width, patch_size[1], n_channels)

    # Swap the first and second axes of the grid to revert the stacking order
    grid = jnp.swapaxes(grid, 1, 2)

    # Reshape the grid of tiles into the original list of tiles
    patches = grid.reshape(height * width, patch_size[0], patch_size[1], n_channels)

    return patches


def unflatten_patches(patches: Array, image_size: Tuple[int, int]) -> Array:
    image_height = image_size[0]
    image_width = image_size[1]
    patch_height = patches.shape[1]
    patch_width = patches.shape[2]
    n_channels = patches.shape[3]

    # Reshape the list of tiles into a 2D grid
    grid = patches.reshape(
        image_height // patch_height,
        image_width // patch_width,
        patch_height,
        patch_width,
        n_channels,
    )

    # Swap the first and second axes of the grid to change the order of stacking
    grid = jnp.swapaxes(grid, 1, 2)

    # Reshape and stack the grid tiles horizontally and vertically to form the final image
    image = grid.reshape(image_height, image_width, n_channels)

    return image



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
import jax

import jax.numpy as jnp
from jax import Array

from . import graphics
from .components import DISCARD_PILE_IDX
from .entities import State
from .grid import align, idx_from_coordinates, crop, view_cone


RADIUS = 3


def none(
    state: State,
) -> Array:
    return jnp.asarray(())


def categorical(
    state: State,
) -> Array:
    # get idx of entity on the set of patches
    indices = idx_from_coordinates(state.grid, state.get_positions())
    # get tags corresponding to the entities
    tags = state.get_tags()
    # set tags on the flat set of patches
    shape = state.grid.shape
    grid = state.grid.reshape(-1).at[indices].set(tags)
    # unflatten patches to reconstruct the grid
    return grid.reshape(shape)


def categorical_first_person(
    state: State,
) -> Array:
    # get transparency map
    transparency_map = jnp.where(state.grid == 0, 1, 0)
    positions = state.get_positions()
    transparent = state.get_transparency()
    transparency_map = transparency_map.at[tuple(positions.T)].set(~transparent)

    # apply view mask
    player = state.get_player()
    view = view_cone(transparency_map, player.position, RADIUS)

    # get categorical representation
    tags = state.get_tags()
    obs = state.grid.at[tuple(positions.T)].set(tags) * view

    # crop grid to agent's view
    obs = crop(obs, player.position, player.direction, RADIUS)

    return obs


def rgb(
    state: State,
) -> Array:
    # for 1-d vs 2-d indexing benchamarks
    # see https://github.com/epignatelli/navix/tree/observation/2dindexing

    # get idx of entity on the flat set of patches
    indices = idx_from_coordinates(state.grid, state.get_positions())
    # get tiles corresponding to the entities
    tiles = state.get_sprites()
    # set tiles on the flat set of patches
    patches = state.cache.patches.at[indices].set(tiles)
    # remove discard pile
    patches = patches[:DISCARD_PILE_IDX]
    # unflatten patches to reconstruct the image
    image_size = (
        state.grid.shape[0] * graphics.TILE_SIZE,
        state.grid.shape[1] * graphics.TILE_SIZE,
    )
    image = graphics.unflatten_patches(patches, image_size)
    return image


def rgb_first_person(
    state: State,
) -> Array:
    # calculate final image size
    image_size = (
        state.grid.shape[0] * graphics.TILE_SIZE,
        state.grid.shape[1] * graphics.TILE_SIZE,
    )

    # get agent's view
    transparency_map = jnp.where(state.grid == 0, 1, 0)
    positions = state.get_positions()
    transparent = state.get_transparency()
    transparency_map = transparency_map.at[tuple(positions.T)].set(~transparent)
    player = state.get_player()
    view = view_cone(transparency_map, player.position, RADIUS)
    view = jax.image.resize(view, image_size, method="nearest")
    view = jnp.tile(view[..., None], (1, 1, 3))

    # get sprites aligned to player's direction
    sprites = state.get_sprites()
    sprites = jax.vmap(lambda x: align(x, jnp.asarray(0), player.direction))(sprites)

    # align sprites to player's direction
    indices = idx_from_coordinates(state.grid, state.get_positions())
    patches = state.cache.patches.at[indices].set(sprites)

    # remove discard pile
    patches = patches[:DISCARD_PILE_IDX]
    # rearrange the sprites in a grid
    patchwork = patches.reshape(*state.grid.shape, *patches.shape[1:])

    # crop grid to agent's view
    patchwork = crop(patchwork, player.position, player.direction, RADIUS)

    # reconstruct image
    obs = jnp.swapaxes(patchwork, 1, 2)
    shape = obs.shape
    obs = obs.reshape(shape[0] * shape[1], shape[2] * shape[3], *shape[4:])
    return obs



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import jax
import jax.numpy as jnp
from jax import Array

from .entities import Entities, State
from .components import DISCARD_PILE_COORDS
from .grid import translate, rotate, positions_equal
from jax_enums import Enumerable as Enum


class Directions(Enum):
    EAST = 0
    SOUTH = 1
    WEST = 2
    NORTH = 3


def _rotate(state: State, spin: int) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)

    # update player's direction
    direction = rotate(player.direction, spin)

    # update sprite representation
    player = player.replace(direction=direction)

    state = state.set_player(player)

    return state


def _walkable(state: State, position: Array) -> Array:
    # according to the grid
    walkable = jnp.equal(state.grid[tuple(position)], 0)

    for k in state.entities:
        obstructs = jnp.logical_and(
            jnp.logical_not(state.entities[k].walkable),
            positions_equal(state.entities[k].position, position),
        )
        walkable = jnp.logical_and(walkable, jnp.any(jnp.logical_not(obstructs)))
    return jnp.asarray(walkable, dtype=jnp.bool_)


def _move(state: State, direction: Array) -> State:
    if Entities.PLAYER not in state.entities:
        return state

    player = state.get_player(idx=0)
    new_position = translate(player.position, direction)
    can_move = _walkable(state, new_position)
    new_position = jnp.where(can_move, new_position, player.position)
    player = player.replace(position=new_position)
    state = state.set_player(player)
    return state


def undefined(state: State) -> State:
    # this is problematic because jax.lax.switch evaluates
    # all *python* branches (no XLA computation is performed)
    # even though only one is selected
    # one option is the following, but this breaks type checking
    # def raise_error(state: State) -> State:
    #     raise ValueError("Undefined action")
    # jax.debug.callback(raise_error)
    raise ValueError("Undefined action")


def noop(state: State) -> State:
    return state


def rotate_cw(state: State) -> State:
    return _rotate(state, 1)


def rotate_ccw(state: State) -> State:
    return _rotate(state, -1)


def forward(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction)


def right(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 1)


def backward(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 2)


def left(state: State) -> State:
    player = state.get_player(idx=0)
    return _move(state, player.direction + 3)


def pickup(state: State) -> State:
    if Entities.KEY not in state.entities:
        return state

    player = state.get_player(idx=0)
    keys = state.get_keys()

    position_in_front = translate(player.position, player.direction)

    key_found = positions_equal(position_in_front, keys.position)

    # update keys
    positions = jnp.where(key_found, DISCARD_PILE_COORDS, keys.position)
    keys = keys.replace(position=positions)

    # update player's pocket, if the pocket has something else, we overwrite it
    key = jnp.sum(keys.id * key_found, dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(key_found), lambda: player.replace(pocket=key), lambda: player
    )

    state = state.set_player(player)
    state = state.set_keys(keys)
    return state


def open(state: State) -> State:
    """Unlocks and opens an openable object (like a door) if possible"""

    if Entities.DOOR not in state.entities:
        return state

    # get the tile in front of the player
    player = state.get_player(idx=0)
    doors = state.get_doors()

    position_in_front = translate(player.position, player.direction)

    # check if there is a door in front of the player
    door_found = positions_equal(position_in_front, doors.position)

    # and that, if so, either it does not require a key or the player has the key
    requires_key = doors.requires != -1
    key_match = player.pocket == doors.requires
    can_open = door_found & (key_match | ~requires_key)

    # update doors if closed and can_open
    do_open = ~doors.open & can_open
    open = jnp.where(do_open, True, doors.open)
    doors = doors.replace(open=open)

    # remove key from player's pocket
    pocket = jnp.asarray(player.pocket * jnp.any(can_open), dtype=jnp.int32)
    player = jax.lax.cond(
        jnp.any(can_open), lambda: player.replace(pocket=pocket), lambda: player
    )

    state = state.set_player(player)
    state = state.set_doors(doors)

    return state


# TODO(epignatelli): a mutable dictionary here is dangerous
ACTIONS = {
    # -1: undefined,
    0: noop,
    1: rotate_cw,
    2: rotate_ccw,
    3: forward,
    4: right,
    5: backward,
    6: left,
    7: pickup,
    8: open,
}
